@@ -1,46 +1,29 @@
 /*
- 
-              Copyright (C) 1997-1998 Hewlett-Packard Company
-                         ALL RIGHTS RESERVED.
- 
-  The enclosed software and documentation includes copyrighted works
-  of Hewlett-Packard Co. For as long as you comply with the following
-  limitations, you are hereby authorized to (i) use, reproduce, and
-  modify the software and documentation, and to (ii) distribute the
-  software and documentation, including modifications, for
-  non-commercial purposes only.
-      
-  1.  The enclosed software and documentation is made available at no
-      charge in order to advance the general development of
-      high-performance networking and computing products.
- 
-  2.  You may not delete any copyright notices contained in the
-      software or documentation. All hard copies, and copies in
-      source code or object code form, of the software or
-      documentation (including modifications) must contain at least
-      one of the copyright notices.
- 
-  3.  The enclosed software and documentation has not been subjected
-      to testing and quality control and is not a Hewlett-Packard Co.
-      product. At a future time, Hewlett-Packard Co. may or may not
-      offer a version of the software and documentation as a product.
-  
-  4.  THE SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS".
-      HEWLETT-PACKARD COMPANY DOES NOT WARRANT THAT THE USE,
-      REPRODUCTION, MODIFICATION OR DISTRIBUTION OF THE SOFTWARE OR
-      DOCUMENTATION WILL NOT INFRINGE A THIRD PARTY'S INTELLECTUAL
-      PROPERTY RIGHTS. HP DOES NOT WARRANT THAT THE SOFTWARE OR
-      DOCUMENTATION IS ERROR FREE. HP DISCLAIMS ALL WARRANTIES,
-      EXPRESS AND IMPLIED, WITH REGARD TO THE SOFTWARE AND THE
-      DOCUMENTATION. HP SPECIFICALLY DISCLAIMS ALL WARRANTIES OF
-      MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-  
-  5.  HEWLETT-PACKARD COMPANY WILL NOT IN ANY EVENT BE LIABLE FOR ANY
-      DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
-      (INCLUDING LOST PROFITS) RELATED TO ANY USE, REPRODUCTION,
-      MODIFICATION, OR DISTRIBUTION OF THE SOFTWARE OR DOCUMENTATION.
- 
+    httperf -- a tool for measuring web server performance
+    Copyright (C) 2000  Hewlett-Packard Company
+    Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
+
+    This file is part of httperf, a web server performance measurment
+    tool.
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License as
+    published by the Free Software Foundation; either version 2 of the
+    License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+    02111-1307 USA
 */
+
+#include "config.h"
+
 #include <assert.h>
 #include <ctype.h>
 #include <errno.h>
@@ -89,6 +72,11 @@ static char http10req[] =
 static char http11req[] =
   " HTTP/1.1\r\nUser-Agent: httperf/"VERSION"\r\nHost: ";
 
+static char http10req_nohost[] =
+  " HTTP/1.0\r\nUser-Agent: httperf/"VERSION"\r\n";
+static char http11req_nohost[] =
+  " HTTP/1.1\r\nUser-Agent: httperf/"VERSION"\r\n";
+
 #ifndef SOL_TCP
 # define SOL_TCP 6	/* probably ought to do getprotlbyname () */
 #endif
@@ -112,12 +100,14 @@ static char http11req[] =
   enum Syscalls
     {
       SC_BIND, SC_CONNECT, SC_READ, SC_SELECT, SC_SOCKET, SC_WRITEV,
+      SC_SSL_READ, SC_SSL_WRITEV,
       SC_NUM_SYSCALLS
     };
 
   static const char * const syscall_name[SC_NUM_SYSCALLS] =
     {
-      "bind", "connct", "read", "select", "socket", "writev"
+      "bind", "connct", "read", "select", "socket", "writev",
+      "ssl_read", "ssl_writev"
     };
   static Time syscall_time[SC_NUM_SYSCALLS];
   static u_int syscall_count[SC_NUM_SYSCALLS];
@@ -215,12 +205,14 @@ lffs (long w)
   else
     {
       r = ffs (w);
+#if SIZEOF_LONG > 4
       if (r == 0)
 	{
 	  r = ffs (w >> (8*sizeof (int)));
 	  if (r > 0)
 	    r += 8*sizeof (int);
 	}
+#endif
     }
   return r;
 }
@@ -355,26 +347,45 @@ set_active (Conn *s, fd_set *fdset)
 }
 
 static void
-do_send (Conn *s)
+do_send (Conn *conn)
 {
   int async_errno, len;
   struct iovec *iovp;
-  int sd = s->sd;
-  ssize_t nsent;
+  int sd = conn->sd;
+  ssize_t nsent = 0;
   Any_Type arg;
-  Call *c;
+  Call *call;
 
   while (1)
     {
-      c = s->sendq;
-      assert (c);
-      SYSCALL (WRITEV,
-	       nsent = writev (sd, c->req.iov + c->req.iov_index,
-			       NELEMS (c->req.iov) - c->req.iov_index));
+      call = conn->sendq;
+      assert (call);
+
+      arg.l = 0;
+      event_signal (EV_CALL_SEND_RAW_DATA, (Object *) call, arg);
+
+#ifdef HAVE_SSL
+      if (param.use_ssl)
+	{
+	  extern ssize_t SSL_writev (SSL *, const struct iovec *, int);
+	  SYSCALL (SSL_WRITEV,
+		   nsent = SSL_writev(conn->ssl,
+				      call->req.iov + call->req.iov_index,
+				      (NELEMS (call->req.iov)
+				       - call->req.iov_index)));
+	}
+      else
+#endif
+	{
+	  SYSCALL (WRITEV,
+		   nsent = writev (sd, call->req.iov + call->req.iov_index,
+				   (NELEMS (call->req.iov)
+				    - call->req.iov_index)));
+	}
 
       if (DBG > 0)
-	fprintf (stderr, "do_send.%lu: wrote %ld bytes\n", c->id,
-		 (long) nsent);
+	fprintf (stderr, "do_send.%lu: wrote %ld bytes on %p\n", call->id,
+		 (long) nsent, conn);
 
       if (nsent < 0)
 	{
@@ -390,14 +401,14 @@ do_send (Conn *s)
 	    fprintf (stderr, "%s.do_send: writev() failed: %s\n",
 		     prog_name, strerror (errno));
 
-	  conn_failure (s, errno);
+	  conn_failure (conn, errno);
 	  return;
 	}
 
-      c->req.size += nsent;
+      call->req.size += nsent;
 
-      iovp = c->req.iov + c->req.iov_index;
-      while (iovp < c->req.iov + NELEMS (c->req.iov))
+      iovp = call->req.iov + call->req.iov_index;
+      while (iovp < call->req.iov + NELEMS (call->req.iov))
 	{
 	  if (nsent < iovp->iov_len)
 	    {
@@ -409,57 +420,59 @@ do_send (Conn *s)
 	    {
 	      /* we're done with this fragment: */
 	      nsent -= iovp->iov_len;
-	      *iovp = c->req.iov_saved;
+	      *iovp = call->req.iov_saved;
 	      ++iovp;
-	      c->req.iov_saved = *iovp;
+	      call->req.iov_saved = *iovp;
 	    }
 	}
-      c->req.iov_index = iovp - c->req.iov;
-      if (c->req.iov_index < NELEMS (c->req.iov))
+      call->req.iov_index = iovp - call->req.iov;
+      if (call->req.iov_index < NELEMS (call->req.iov))
 	{
 	  /* there are more header bytes to write */
-	  c->timeout = param.timeout ? timer_now () + param.timeout : 0.0;
-	  set_active (s, &wrfds);
+	  call->timeout = param.timeout ? timer_now () + param.timeout : 0.0;
+	  set_active (conn, &wrfds);
 	  return;
 	}
 
       /* we're done with sending this request */
-      s->sendq = c->sendq_next;
-      if (!s->sendq)
+      conn->sendq = call->sendq_next;
+      if (!conn->sendq)
 	{
-	  s->sendq_tail = 0;
+	  conn->sendq_tail = 0;
 	  FD_CLR (sd, &wrfds);
 	}
       arg.l = 0;
-      event_signal (EV_CALL_SEND_STOP, (Object *) c, arg);
-      if (s->state >= S_CLOSING)
+      event_signal (EV_CALL_SEND_STOP, (Object *) call, arg);
+      if (conn->state >= S_CLOSING)
 	{
-	  call_dec_ref (c);
+	  call_dec_ref (call);
 	  return;
 	}
 
       /* get ready to receive matching reply (note that we implicitly
          pass on the reference to the call from the sendq to the
          recvq): */
-      if (!s->recvq)
-	s->recvq = s->recvq_tail = c;
+      call->recvq_next = 0;
+      if (!conn->recvq)
+	conn->recvq = conn->recvq_tail = call;
       else
 	{
-	  s->recvq_tail->recvq_next = c;
-	  s->recvq_tail = c;
+	  conn->recvq_tail->recvq_next = call;
+	  conn->recvq_tail = call;
 	}
-      c->timeout = param.timeout + param.think_timeout;
-      if (c->timeout > 0.0)
-	c->timeout += timer_now ();
-      set_active (s, &rdfds);
-      s->state = S_REPLY_STATUS;	/* expecting reply status */
+      call->timeout = param.timeout + param.think_timeout;
+      if (call->timeout > 0.0)
+	call->timeout += timer_now ();
+      set_active (conn, &rdfds);
+      if (conn->state < S_REPLY_STATUS)
+	conn->state = S_REPLY_STATUS;	/* expecting reply status */
 
-      if (!s->sendq)
+      if (!conn->sendq)
 	return;
 
       arg.l = 0;
-      event_signal (EV_CALL_SEND_START, (Object *) s->sendq, arg);
-      if (s->state >= S_CLOSING)
+      event_signal (EV_CALL_SEND_START, (Object *) conn->sendq, arg);
+      if (conn->state >= S_CLOSING)
 	return;
     }
 }
@@ -488,33 +501,50 @@ do_recv (Conn *s)
 {
   char *cp, buf[8193];
   Call *c = s->recvq;
+  int i, saved_errno;
+  ssize_t nread = 0;
   size_t buf_len;
-  ssize_t nread;
-  int i;
 
   assert (c);
 
-  SYSCALL (READ,
-	   nread = read (s->sd, buf, sizeof (buf) - 1));
+#ifdef HAVE_SSL
+  if (param.use_ssl)
+    {
+      SYSCALL (SSL_READ,
+	       nread = SSL_read (s->ssl, buf, sizeof (buf) - 1));
+    }
+  else
+#endif
+    {
+      SYSCALL (READ,
+	       nread = read (s->sd, buf, sizeof (buf) - 1));
+    }
+  saved_errno = errno;
   if (nread <= 0)
     {
       if (DBG > 0)
 	{
-	  fprintf (stderr, "do_recv.%lu: received %lu reply bytes\n", c->id,
-		   (u_long) (c->reply.header_bytes + c->reply.content_bytes));
+	  fprintf (stderr, "do_recv.%lu: received %lu reply bytes on %p\n",
+		   c->id,
+		   (u_long) (c->reply.header_bytes + c->reply.content_bytes),
+		   s);
 	  if (nread < 0)
 	    fprintf (stderr, "%s.do_recv: read() failed: %s\n",
-		     prog_name, strerror (errno));
+		     prog_name, strerror (saved_errno));
 	}
       if (nread < 0)
 	{
-	  if (errno != EAGAIN)
-	    conn_failure (s, errno);
+	  if (saved_errno != EAGAIN)
+	    conn_failure (s, saved_errno);
 	}
-      else if (nread == 0 && s->state == S_REPLY_STATUS)
+      else if (s->state != S_REPLY_DATA)
 	conn_failure (s, ECONNRESET);
       else
-	recv_done (c);
+	{
+	  if (s->state < S_CLOSING)
+	    s->state = S_REPLY_DONE;
+	  recv_done (c);
+	}
       return;
     }
   buf[nread] = '\0';	/* ensure buffer is '\0' terminated */
@@ -641,6 +671,14 @@ core_init (void)
 	       prog_name, strerror (errno));
       exit (1);
     }
+
+  if (rlimit.rlim_max > FD_SETSIZE)
+    {
+      fprintf (stderr, "%s: warning: open file limit > FD_SETSIZE; "
+	       "limiting max. # of open files to FD_SETSIZE\n", prog_name);
+      rlimit.rlim_max = FD_SETSIZE;
+    }
+
   rlimit.rlim_cur = rlimit.rlim_max;
   if (setrlimit (RLIMIT_NOFILE, &rlimit) < 0)
     {
@@ -657,6 +695,76 @@ core_init (void)
     core_addr_intern (param.server, strlen (param.server), param.port);
 }
 
+#ifdef HAVE_SSL
+
+void
+core_ssl_connect (Conn *s)
+{
+  Any_Type arg;
+  int ssl_err;
+
+  if (DBG > 2)
+    fprintf (stderr, "core_ssl_connect(conn=%p)\n", (void *) s);
+
+  if (SSL_set_fd (s->ssl, s->sd) == 0)
+    {
+      ERR_print_errors_fp (stderr);
+      exit (-1);
+    }
+
+  ssl_err = SSL_connect (s->ssl);
+  if (ssl_err < 0)
+    {
+      int reason = SSL_get_error(s->ssl, ssl_err);
+
+      if (reason == SSL_ERROR_WANT_READ || reason == SSL_ERROR_WANT_WRITE)
+	{
+	  if (DBG > 2)
+	    fprintf (stderr, "core_ssl_connect: want to %s more...\n",
+		     (reason == SSL_ERROR_WANT_READ) ? "read" : "write");
+	  if (reason == SSL_ERROR_WANT_READ && !FD_ISSET (s->sd, &rdfds))
+	    {
+	      FD_CLR (s->sd, &wrfds);
+	      set_active (s, &rdfds);
+	    }
+	  else if (reason == SSL_ERROR_WANT_WRITE && !FD_ISSET (s->sd, &wrfds))
+	    {
+	      FD_CLR (s->sd, &rdfds);
+	      set_active (s, &wrfds);
+	    }
+	  return;
+	}
+      fprintf (stderr,
+	       "%s: failed to connect to SSL server (err=%d, reason=%d)\n",
+	       prog_name, ssl_err, reason);
+      ERR_print_errors_fp (stderr);
+      exit (-1);
+    }
+
+  s->state = S_CONNECTED;
+
+  if (DBG > 0)
+    fprintf (stderr, "core_ssl_connect: SSL is connected!\n");
+
+  if (DBG > 1)
+    {
+      SSL_CIPHER *ssl_cipher;
+
+      ssl_cipher = SSL_get_current_cipher (s->ssl);
+      if (!ssl_cipher)
+	fprintf (stderr, "core_ssl_connect: server refused all client cipher "
+		 "suites!\n");
+      else
+	fprintf (stderr, "core_ssl_connect: cipher=%s, valid=%d, id=%lu\n",
+		 ssl_cipher->name, ssl_cipher->valid, ssl_cipher->id);
+    }
+
+  arg.l = 0;
+  event_signal (EV_CONN_CONNECTED, (Object *) s, arg);
+}
+
+#endif /* HAVE_SSL */
+
 int
 core_connect (Conn *s)
 {
@@ -800,9 +908,16 @@ core_connect (Conn *s)
 	   result = connect (sd, sin, sizeof (*sin)));
   if (result == 0)
     {
-      s->state = S_CONNECTED;
-      arg.l = 0;
-      event_signal (EV_CONN_CONNECTED, (Object *) s, arg);
+#ifdef HAVE_SSL
+      if (param.use_ssl)
+	  core_ssl_connect (s);
+      else
+#endif
+	{
+	  s->state = S_CONNECTED;
+	  arg.l = 0;
+	  event_signal (EV_CONN_CONNECTED, (Object *) s, arg);
+	}
     }
   else if (errno == EINPROGRESS)
     {
@@ -847,7 +962,12 @@ core_send (Conn *conn, Call *call)
 
   call->conn = conn;	/* NO refcounting here (see call.h).  */
 
-  if (!call->req.iov[IE_HOST].iov_base)
+  if (param.no_host_hdr)
+    {
+      call->req.iov[IE_HOST].iov_base = (caddr_t) "";
+      call->req.iov[IE_HOST].iov_len = 0;
+    }
+  else if (!call->req.iov[IE_HOST].iov_base)
     {
       /* Default call's hostname to connection's hostname: */
       call->req.iov[IE_HOST].iov_base = (caddr_t) conn->hostname;
@@ -862,13 +982,29 @@ core_send (Conn *conn, Call *call)
   switch (call->req.version)
     {
     case 0x10000:
-      call->req.iov[IE_PROTL].iov_base = (caddr_t) http10req;
-      call->req.iov[IE_PROTL].iov_len = sizeof (http10req) - 1;
+      if (param.no_host_hdr)
+	{
+	  call->req.iov[IE_PROTL].iov_base = (caddr_t) http10req_nohost;
+	  call->req.iov[IE_PROTL].iov_len = sizeof (http10req_nohost) - 1;
+	}
+      else
+	{
+	  call->req.iov[IE_PROTL].iov_base = (caddr_t) http10req;
+	  call->req.iov[IE_PROTL].iov_len = sizeof (http10req) - 1;
+	}
       break;
 
     case 0x10001:
-      call->req.iov[IE_PROTL].iov_base = http11req;
-      call->req.iov[IE_PROTL].iov_len = sizeof (http11req) - 1;
+      if (param.no_host_hdr)
+	{
+	  call->req.iov[IE_PROTL].iov_base = http11req_nohost;
+	  call->req.iov[IE_PROTL].iov_len = sizeof (http11req_nohost) - 1;
+	}
+      else
+	{
+	  call->req.iov[IE_PROTL].iov_base = http11req;
+	  call->req.iov[IE_PROTL].iov_len = sizeof (http11req) - 1;
+	}
       break;
 
     default:
@@ -920,7 +1056,7 @@ core_close (Conn *conn)
       conn->watchdog = 0;
     }
 
-  /* first, get rid of call pending calls: */
+  /* first, get rid of all pending calls: */
   for (call = conn->sendq; call; call = call_next)
     {
       call_next = call->sendq_next;
@@ -942,6 +1078,11 @@ core_close (Conn *conn)
   event_signal (EV_CONN_CLOSE, (Object *) conn, arg);
   assert (conn->state == S_CLOSING);
 
+#ifdef HAVE_SSL
+  if (param.use_ssl)
+    SSL_shutdown (conn->ssl);
+#endif
+
   if (sd >= 0)
     {
       close (sd);
@@ -1032,22 +1173,29 @@ core_loop (void)
 			  timer_cancel (conn->watchdog);
 			  conn->watchdog = 0;
 			}
-		      if (is_writable)
+		      if (conn->state == S_CONNECTING)
 			{
-			  if (conn->state == S_CONNECTING)
-			    {
-			      conn->state = S_CONNECTED;
-			      FD_CLR (sd, &wrfds);
-
-			      arg.l = 0;
-			      event_signal (EV_CONN_CONNECTED, (Object *) conn,
-					    arg);
-			    }
-			  else if (conn->sendq)
+#ifdef HAVE_SSL
+			  if (param.use_ssl)
+			    core_ssl_connect (conn);
+			  else
+#endif
+			    if (is_writable)
+			      {
+				FD_CLR (sd, &wrfds);
+				conn->state = S_CONNECTED;
+				arg.l = 0;
+				event_signal (EV_CONN_CONNECTED,
+					      (Object *) conn, arg);
+			      }
+			}
+		      else
+			{
+			  if (is_writable && conn->sendq)
 			    do_send (conn);
+			  if (is_readable && conn->recvq)
+			    do_recv (conn);
 			}
-		      if (is_readable && conn->recvq)
-			do_recv (conn);
 
 		      conn_dec_ref (conn);
 
