@@ -1,46 +1,27 @@
 /*
- 
-              Copyright (C) 1997, 1998 Hewlett-Packard Company
-                         ALL RIGHTS RESERVED.
- 
-  The enclosed software and documentation includes copyrighted works
-  of Hewlett-Packard Co. For as long as you comply with the following
-  limitations, you are hereby authorized to (i) use, reproduce, and
-  modify the software and documentation, and to (ii) distribute the
-  software and documentation, including modifications, for
-  non-commercial purposes only.
-      
-  1.  The enclosed software and documentation is made available at no
-      charge in order to advance the general development of
-      high-performance networking and computing products.
- 
-  2.  You may not delete any copyright notices contained in the
-      software or documentation. All hard copies, and copies in
-      source code or object code form, of the software or
-      documentation (including modifications) must contain at least
-      one of the copyright notices.
- 
-  3.  The enclosed software and documentation has not been subjected
-      to testing and quality control and is not a Hewlett-Packard Co.
-      product. At a future time, Hewlett-Packard Co. may or may not
-      offer a version of the software and documentation as a product.
-  
-  4.  THE SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS".
-      HEWLETT-PACKARD COMPANY DOES NOT WARRANT THAT THE USE,
-      REPRODUCTION, MODIFICATION OR DISTRIBUTION OF THE SOFTWARE OR
-      DOCUMENTATION WILL NOT INFRINGE A THIRD PARTY'S INTELLECTUAL
-      PROPERTY RIGHTS. HP DOES NOT WARRANT THAT THE SOFTWARE OR
-      DOCUMENTATION IS ERROR FREE. HP DISCLAIMS ALL WARRANTIES,
-      EXPRESS AND IMPLIED, WITH REGARD TO THE SOFTWARE AND THE
-      DOCUMENTATION. HP SPECIFICALLY DISCLAIMS ALL WARRANTIES OF
-      MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-  
-  5.  HEWLETT-PACKARD COMPANY WILL NOT IN ANY EVENT BE LIABLE FOR ANY
-      DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
-      (INCLUDING LOST PROFITS) RELATED TO ANY USE, REPRODUCTION,
-      MODIFICATION, OR DISTRIBUTION OF THE SOFTWARE OR DOCUMENTATION.
- 
+    httperf -- a tool for measuring web server performance
+    Copyright (C) 2000  Hewlett-Packard Company
+    Contributed by David Mosberger-Tang <davidm@hpl.hp.com>
+
+    This file is part of httperf, a web server performance measurment
+    tool.
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License as
+    published by the Free Software Foundation; either version 2 of the
+    License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+    02111-1307 USA
 */
+
 /*
   Fundamentals:
 
@@ -66,14 +47,18 @@
 	  to fix up this case---simply declare defeat and abort the
 	  test.
  */
+#include "config.h"
+
 #include <ctype.h>
 #include <errno.h>
 #include <getopt.h>
+#include <signal.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+
 #include <sys/time.h>
 #include <sys/resource.h>
 
@@ -83,6 +68,10 @@
 #include <conn.h>
 #include <timer.h>
 
+#ifdef HAVE_SSL
+#  include <openssl/rand.h>
+#endif
+
 #define RATE_INTERVAL	5.0
 
 const char *prog_name;
@@ -94,6 +83,10 @@ struct rusage test_rusage_start;
 struct rusage test_rusage_stop;
 size_t object_type_size[OBJ_NUM_TYPES];
 
+#ifdef HAVE_SSL
+  SSL_CTX *ssl_ctx;
+#endif
+
 #ifdef DEBUG
 int debug_level;
 #endif
@@ -102,6 +95,7 @@ static Time perf_sample_start;
 
 static struct option longopts[] =
   {
+    {"add-header",   required_argument, (int *) &param.additional_header, 0},
     {"burst-length", required_argument, &param.burst_len,		0},
     {"client",	     required_argument, (int *) &param.client,		0},
     {"close-with-reset", no_argument,	&param.close_with_reset,	1},
@@ -112,16 +106,26 @@ static struct option longopts[] =
     {"http-version", required_argument, (int *) &param.http_version,	0},
     {"max-connections", required_argument, &param.max_conns,		0},
     {"max-piped-calls", required_argument, &param.max_piped,		0},
+    {"method",	     required_argument,	(int *) &param.method,		0},
+    {"no-host-hdr",  no_argument,	&param.no_host_hdr,		1},
     {"num-calls",    required_argument, (int *) &param.num_calls,	0},
     {"num-conns",    required_argument, (int *) &param.num_conns,	0},
-    {"port",	     required_argument, (int *) &param.port,		0},
     {"period",	     required_argument,	(int *) &param.rate.mean_iat,	0},
-    {"print-replies", no_argument,	&param.print_replies,		1},
+    {"port",	     required_argument, (int *) &param.port,		0},
+    {"print-reply",  optional_argument, &param.print_reply,		0},
+    {"print-request",optional_argument, &param.print_request, 		0},
     {"rate",	     required_argument,	(int *) &param.rate,		0},
     {"recv-buffer",  required_argument, (int *) &param.recv_buffer_size, 0},
+    {"retry-on-failure", no_argument,	&param.retry_on_failure,	1},
     {"send-buffer",  required_argument, (int *) &param.send_buffer_size, 0},
     {"server",	     required_argument, (int *) &param.server,		0},
+    {"server-name",  required_argument, (int *) &param.server_name,	0},
     {"session-cookies", no_argument,	(int *) &param.session_cookies,	1},
+#ifdef HAVE_SSL
+    {"ssl",	     no_argument,	&param.use_ssl,			1},
+    {"ssl-ciphers",  required_argument, (int *) &param.ssl_cipher_list, 0},
+    {"ssl-no-reuse", no_argument,	&param.ssl_reuse,		0},
+#endif
     {"think-timeout",required_argument, (int *) &param.think_timeout,	0},
     {"timeout",      required_argument,	(int *) &param.timeout,		0},
     {"uri",	     required_argument, (int *) &param.uri,		0},
@@ -130,6 +134,7 @@ static struct option longopts[] =
     {"wlog",	     required_argument, (int *) &param.wlog,            0},
     {"wsess",	     required_argument, (int *) &param.wsess,		0},
     {"wsesslog",     required_argument, (int *) &param.wsesslog,	0},
+    {"wsesspage",    required_argument, (int *) &param.wsesspage,	0},
     {"wset",	     required_argument, (int *) &param.wset,		0},
     {0,		     0,			0,				0}
   };
@@ -138,17 +143,23 @@ static void
 usage (void)
 {
   printf ("Usage: %s "
-	  "[-hdvV] [--close-with-reset] [--debug N] [--help] [--hog]\n"
-	  "\t[--verbose] [--version] [--print-replies]\n"
-	  "\t[--client N/N] [--server S] [--port N] [--uri S] "
-	  "[--num-calls N]\n"
-	  "\t[--burst-length N] [--failure-status N] [--http-version S]\n"
-	  "\t[--max-connections N] [--max-piped-calls N]\n"
-	  "\t[--num-conns N] [--rate X] [--period [d|u|e]T1[,T2]\n"
-	  "\t[--timeout X] [--think-timeout X]\n"
-	  "\t[--recv-buffer N] [--send-buffer N] [--session-cookies]\n"
-	  "\t[--wlog y|n,file] [--wsess N,N,X] [--wsesslog N,X,file] "
-	  "[--wset N,X]\n",
+	  "[-hdvV] [--add-header S] [--burst-length N] [--client N/N]\n"
+	  "\t[--close-with-reset] [--debug N] [--failure-status N]\n"
+	  "\t[--help] [--hog] [--http-version S] [--max-connections N]\n"
+	  "\t[--max-piped-calls N] [--method S] [--no-host-hdr]\n"
+	  "\t[--num-calls N] [--num-conns N] [--period [d|u|e]T1[,T2]]\n"
+	  "\t[--port N] "
+	  "[--print-reply [header|body]] [--print-request [header|body]]\n"
+	  "\t[--rate X] [--recv-buffer N] [--retry-on-failure] "
+	  "[--send-buffer N]\n"
+	  "\t[--server S] [--server-name S] [--session-cookies]\n"
+#ifdef HAVE_SSL
+	  "\t[--ssl] [--ssl-ciphers L] [--ssl-no-reuse]\n"
+#endif
+	  "\t[--think-timeout X] [--timeout X] [--uri S] [--verbose] "
+	  "[--version]\n"
+	  "\t[--wlog y|n,file] [--wsess N,N,X] [--wsesslog N,X,file]\n"
+	  "\t[--wset N,X]\n",
 	  prog_name);
 }
 
@@ -185,12 +196,13 @@ int
 main (int argc, char **argv)
 {
   extern Load_Generator uri_fixed, uri_wlog, uri_wset, conn_rate, call_seq;
-  extern Load_Generator wsess, wsesslog, sess_cookie;
+  extern Load_Generator wsess, wsesslog, wsesspage, sess_cookie, misc;
   extern Stat_Collector stats_basic, session_stat;
   extern Stat_Collector stats_print_reply;
   extern char *optarg;
+  int session_workload = 0;
   int num_gen = 3;
-  Load_Generator *gen[4] =
+  Load_Generator *gen[5] =
     {
       &call_seq,
       &uri_fixed,
@@ -208,6 +220,13 @@ main (int argc, char **argv)
   void *flag;
   Time t;
 
+#ifdef __FreeBSD__
+  /* This works around a bug in earlier versions of FreeBSD that cause
+     non-finite IEEE arithmetic to cause SIGFPE instead of the
+     non-finite arithmetic as defined by IEEE.  */
+  fpsetmask (0);
+#endif
+
   object_type_size[OBJ_CONN] = sizeof (Conn);
   object_type_size[OBJ_CALL] = sizeof (Call);
 
@@ -215,7 +234,7 @@ main (int argc, char **argv)
   param.client.id = 0;
   param.client.num_clients = 1;
   param.server = "localhost";
-  param.port = 80;
+  param.port = -1;
   param.uri = "/";
   param.num_calls = 1;
   param.burst_len = 1;
@@ -225,6 +244,9 @@ main (int argc, char **argv)
   param.send_buffer_size =  4096;
   param.recv_buffer_size = 16384;
   param.rate.dist = DETERMINISTIC;
+#ifdef HAVE_SSL
+  param.ssl_reuse = 1;
+#endif
 
   /* get program name: */
   prog_name = strrchr (argv[0], '/');
@@ -241,7 +263,11 @@ main (int argc, char **argv)
 	case 0:
 	  flag = longopts[longindex].flag;
 
-	  if (flag == &param.num_calls)
+	  if (flag == &param.method)
+	    param.method = optarg;
+	  else if (flag == &param.additional_header)
+	    param.additional_header = optarg;
+	  else if (flag == &param.num_calls)
 	    {
 	      errno = 0;
 	      param.num_calls = strtoul (optarg, &end, 10);
@@ -326,13 +352,28 @@ main (int argc, char **argv)
 	      errno = 0;
 	      param.port = strtoul (optarg, &end, 10);
 	      if (errno == ERANGE || end == optarg || *end
-		  || param.port > 0xffff)
+		  || (unsigned) param.port > 0xffff)
 		{
 		  fprintf (stderr, "%s: illegal port number %s\n",
 			   prog_name, optarg);
 		  exit (1);
 		}
 	    }
+	  else if (flag == &param.print_request || flag == &param.print_reply)
+	    {
+	      int val;
+
+	      if (!optarg)
+		val = PRINT_HEADER | PRINT_BODY;
+	      else
+		switch (tolower (optarg[0]))
+		  {
+		  case 'h': val = PRINT_HEADER;	break;
+		  case 'b': val = PRINT_BODY;	break;
+		  default:  val = PRINT_HEADER | PRINT_BODY; break;
+		  }
+	      *(int *) flag = val;
+	    }
 	  else if (flag == &param.rate)
 	    {
 	      errno = 0;
@@ -474,6 +515,12 @@ main (int argc, char **argv)
 	    }
 	  else if (flag == &param.server)
 	    param.server = optarg;
+	  else if (flag == &param.server_name)
+	    param.server_name = optarg;
+#ifdef HAVE_SSL
+	  else if (flag == &param.ssl_cipher_list)
+	    param.ssl_cipher_list = optarg;
+#endif
 	  else if (flag == &param.uri)
 	    param.uri = optarg;
 	  else if (flag == &param.think_timeout)
@@ -513,7 +560,7 @@ main (int argc, char **argv)
 	      stat[num_stats++] = &session_stat;
 
 	      errno = 0;
-	      name = "bad number of user connections (1st param)";
+	      name = "bad number of sessions (1st param)";
 	      param.wsess.num_sessions = strtoul (optarg, &end, 0);
 	      if (end == optarg || errno == ERANGE)
 		goto bad_wsess_param;
@@ -549,6 +596,53 @@ main (int argc, char **argv)
 		  fputc ('\n', stderr);
 		  exit (1);
 		}
+	      session_workload = 1;
+	    }
+	  else if (flag == &param.wsesspage)
+	    {
+	      num_gen = 2;		/* XXX fix me---somehow */
+	      gen[0] = &wsesspage;
+
+	      stat[num_stats++] = &session_stat;
+
+	      errno = 0;
+	      name = "bad number of sessions (1st param)";
+	      param.wsesspage.num_sessions = strtoul (optarg, &end, 0);
+	      if (end == optarg || errno == ERANGE)
+		goto bad_wsesspage_param;
+	      optarg = end + 1;
+
+	      name = "bad number of user requests per session (2nd param)";
+	      if (*end != ',')
+		goto bad_wsesspage_param;
+	      optarg = end + 1;
+
+	      param.wsesspage.num_reqs = strtoul (optarg, &end, 0);
+	      if (end == optarg || errno == ERANGE)
+		goto bad_wsesspage_param;
+
+	      name = "bad user think time (3rd param)";
+	      if (*end != ',')
+		goto bad_wsesspage_param;
+	      optarg = end + 1;
+
+	      param.wsesspage.think_time = strtod (optarg, &end);
+	      if (end == optarg || errno == ERANGE 
+		  || param.wsesspage.think_time < 0.0)
+		goto bad_wsesspage_param;
+
+	      name = "extraneous parameter";
+	      if (*end)
+		{
+		bad_wsesspage_param:
+		  fprintf (stderr, "%s: %s in --wsesspage arg (rest: `%s')",
+			   prog_name, name, end);
+		  if (errno)
+		    fprintf (stderr, ": %s", strerror (errno));
+		  fputc ('\n', stderr);
+		  exit (1);
+		}
+	      session_workload = 1;
 	    }
 	  else if (flag == &param.wsesslog)
 	    {
@@ -558,7 +652,7 @@ main (int argc, char **argv)
 	      stat[num_stats++] = &session_stat;
 
 	      errno = 0;
-	      name = "bad number of user connections (1st param)";
+	      name = "bad number of sessions (1st param)";
 	      param.wsesslog.num_sessions = strtoul (optarg, &end, 0);
 	      if (end == optarg || errno == ERANGE)
 		goto bad_wsesslog_param;
@@ -600,6 +694,7 @@ main (int argc, char **argv)
  		  fputc ('\n', stderr);
 		  exit (1);
 		}
+	      session_workload = 1;
 	    }
 	  else if (flag == &param.wset)
 	    {
@@ -691,16 +786,66 @@ main (int argc, char **argv)
 	}
     }
 
-  if (param.print_replies)
+#ifdef HAVE_SSL
+  if (param.use_ssl)
+    {
+      char buf[1024];
+
+      if (param.port < 0)
+	param.port = 443;
+
+      SSL_load_error_strings ();
+      SSLeay_add_ssl_algorithms ();
+
+      /* for some strange reason, SSLv23_client_method () doesn't work here */
+      ssl_ctx = SSL_CTX_new (SSLv3_client_method ());
+      if (!ssl_ctx)
+	{
+	  ERR_print_errors_fp (stderr);
+	  exit (-1);
+	}
+
+      memset (buf, 0, sizeof (buf));
+      RAND_seed (buf, sizeof (buf));
+    }
+#endif
+  if (param.port < 0)
+    param.port = 80;
+
+  if (param.print_reply || param.print_request)
     stat[num_stats++] = &stats_print_reply;
 
   if (param.session_cookies)
-    gen[num_gen++] = &sess_cookie;
+    {
+      if (!session_workload)
+	{
+	  fprintf (stderr, "%s: --session-cookie requires session-oriented "
+		   "workload (e.g., --wsess)\n", prog_name);
+	  exit (-1);
+	}
+      gen[num_gen++] = &sess_cookie;
+    }
+
+  if (param.additional_header || param.method)
+    gen[num_gen++] = &misc;
 
   /* echo command invocation for logging purposes: */
   printf ("%s", prog_name);
   if (verbose) printf (" --verbose");
-  if (param.print_replies) printf (" --print-replies");
+  switch (param.print_reply)
+    {
+    case 0:		break;
+    case PRINT_HEADER:	printf (" --print-reply=header"); break;
+    case PRINT_BODY:	printf (" --print-reply=body"); break;
+    default:		printf (" --print-reply"); break;
+    }
+  switch (param.print_request)
+    {
+    case 0:		break;
+    case PRINT_HEADER:	printf (" --print-request=header"); break;
+    case PRINT_BODY:	printf (" --print-request=body"); break;
+    default:		printf (" --print-request"); break;
+    }
   if (param.hog) printf (" --hog");
   if (param.close_with_reset) printf (" --close-with-reset");
   if (param.think_timeout > 0) printf (" --think-timeout=%g",
@@ -708,6 +853,7 @@ main (int argc, char **argv)
   if (param.timeout > 0) printf (" --timeout=%g", param.timeout);
   printf (" --client=%u/%u", param.client.id, param.client.num_clients);
   if (param.server) printf (" --server=%s", param.server);
+  if (param.server_name) printf (" --server_name=%s", param.server_name);
   if (param.port) printf (" --port=%d", param.port);
   if (param.uri) printf (" --uri=%s", param.uri);
   if (param.failure_status) printf (" --failure-status=%u",
@@ -743,9 +889,18 @@ main (int argc, char **argv)
 	}
     }
   printf (" --send-buffer=%d", param.send_buffer_size);
+  if (param.retry_on_failure) printf (" --retry-on-failure");
   printf (" --recv-buffer=%d", param.recv_buffer_size);
-  if (param.session_cookies)
-    printf (" --session-cookies");
+  if (param.session_cookies) printf (" --session-cookies");
+#ifdef HAVE_SSL
+  if (param.use_ssl) printf (" --ssl");
+  if (param.ssl_cipher_list)
+    printf(" --ssl-ciphers=%s", param.ssl_cipher_list);
+  if (!param.ssl_reuse) printf (" --ssl-no-reuse");
+#endif
+  if (param.additional_header)
+    printf (" --add-header='%s'", param.additional_header);
+  if (param.method) printf (" --method=%s", param.method);
   if (param.wsesslog.num_sessions)
     {
       /* This overrides any --wsess, --num-conns, --num-calls,
@@ -753,6 +908,11 @@ main (int argc, char **argv)
       printf (" --wsesslog=%u,%.3f,%s", param.wsesslog.num_sessions,
 	      param.wsesslog.think_time, param.wsesslog.file);
     }
+  else if (param.wsesspage.num_sessions)
+    {
+      printf (" --wsesspage=%u,%u,%.3f", param.wsesspage.num_sessions,
+	      param.wsesspage.num_reqs, param.wsesspage.think_time);
+    }
   else
     {
       if (param.wsess.num_sessions)
@@ -774,6 +934,8 @@ main (int argc, char **argv)
   timer_init ();
   core_init ();
 
+  signal (SIGINT, (void (*)()) core_exit);
+
   for (i = 0; i < num_stats; ++i)
     (*stat[i]->init)();
   for (i = 0; i < num_gen; ++i)
